@@ -16,6 +16,7 @@ from beartype import beartype
 from icontract import ensure, require
 
 from specfact_cli.agents.base import AgentMode
+from specfact_cli.migrations.plan_migrator import get_current_schema_version
 from specfact_cli.models.plan import Idea, Metadata, PlanBundle, Product
 
 
@@ -112,12 +113,12 @@ Use your AI capabilities to understand the codebase:
 
 2. **Convert to YAML** using proper YAML formatting (2-space indentation, no flow style)
 
-3. **Write to file**: `.specfact/plans/<name>-<timestamp>.bundle.yaml`
+3. **Write to file**: `.specfact/plans/<name>-<timestamp>.bundle.<format>`
    - If no name provided, ask user for a meaningful plan name (e.g., "API Client v2", "User Authentication", "Payment Processing")
    - Name will be automatically sanitized (lowercased, spaces/special chars removed) for filesystem persistence
    - Use ISO 8601 timestamp format: `YYYY-MM-DDTHH-MM-SS`
    - Ensure directory exists: `.specfact/plans/`
-   - Example: `.specfact/plans/api-client-v2.2025-11-04T22-17-22.bundle.yaml`
+   - Example: `.specfact/plans/api-client-v2.2025-11-04T22-17-22.bundle.<format>`
 
 ### Step 3: Present Results
 
@@ -381,11 +382,18 @@ Dependencies: {len(dependencies)} dependency files found
         )
 
         return PlanBundle(
-            version="1.0",
+            version=get_current_schema_version(),
             idea=idea,
             business=None,
             product=product,
             features=[],
-            metadata=Metadata(stage="draft", promoted_at=None, promoted_by=None),
+            metadata=Metadata(
+                stage="draft",
+                promoted_at=None,
+                promoted_by=None,
+                analysis_scope=None,
+                entry_point=None,
+                summary=None,
+            ),
             clarifications=None,
         )
