@@ -16,14 +16,14 @@ import typer
 from beartype import beartype
 from icontract import require
 from rich.console import Console
-from rich.progress import Progress, SpinnerColumn, TextColumn
+from rich.progress import Progress, SpinnerColumn, TextColumn, TimeElapsedColumn
 
 from specfact_cli import runtime
 from specfact_cli.models.bridge import AdapterType
 from specfact_cli.models.plan import Feature, PlanBundle
 from specfact_cli.models.project import BundleManifest, BundleVersions, ProjectBundle
 from specfact_cli.telemetry import telemetry
-from specfact_cli.utils.bundle_loader import save_project_bundle
+from specfact_cli.utils.progress import save_bundle_with_progress
 
 
 app = typer.Typer(help="Import codebases and external tool projects (e.g., Spec-Kit, Linear, Jira) to contract format")
@@ -95,6 +95,7 @@ def _check_incremental_changes(
         with Progress(
             SpinnerColumn(),
             TextColumn("[progress.description]{task.description}"),
+            TimeElapsedColumn(),
             console=console,
         ) as progress:
             task = progress.add_task("[cyan]Checking for changes...", total=None)
@@ -139,21 +140,10 @@ def _check_incremental_changes(
 def _load_existing_bundle(bundle_dir: Path) -> PlanBundle | None:
     """Load existing project bundle and convert to PlanBundle."""
     from specfact_cli.models.plan import PlanBundle as PlanBundleModel
-    from specfact_cli.utils.bundle_loader import load_project_bundle
+    from specfact_cli.utils.progress import load_bundle_with_progress
 
     try:
-        with Progress(
-            SpinnerColumn(),
-            TextColumn("[progress.description]{task.description}"),
-            console=console,
-        ) as progress:
-            task = progress.add_task("[cyan]Loading existing project bundle...", total=None)
-
-            def progress_callback(current: int, total: int, artifact: str) -> None:
-                progress.update(task, description=f"[cyan]Loading artifact {current}/{total}: {artifact}")
-
-            existing_bundle = load_project_bundle(bundle_dir, progress_callback=progress_callback)
-            progress.update(task, description="[green]✓[/green] Bundle loaded")
+        existing_bundle = load_bundle_with_progress(bundle_dir, validate_hashes=False, console_instance=console)
 
         plan_bundle = PlanBundleModel(
             version="1.0",
@@ -253,6 +243,7 @@ def _analyze_codebase(
 
 def _update_source_tracking(plan_bundle: PlanBundle, repo: Path) -> None:
     """Update source tracking with file hashes (parallelized)."""
+    import os
     from concurrent.futures import ThreadPoolExecutor, as_completed
 
     from specfact_cli.utils.source_scanner import SourceArtifactScanner
@@ -275,42 +266,54 @@ def _update_source_tracking(plan_bundle: PlanBundle, repo: Path) -> None:
                 hash_tasks.append((feature, repo / test_file))
 
     if hash_tasks:
-        max_workers = max(1, min(multiprocessing.cpu_count() or 4, 16, len(hash_tasks)))
-        executor = ThreadPoolExecutor(max_workers=max_workers)
-        interrupted = False
-        try:
-            future_to_task = {
-                executor.submit(update_file_hash, feature, file_path): (feature, file_path)
-                for feature, file_path in hash_tasks
-            }
+        import os
+
+        # In test mode, use sequential processing to avoid ThreadPoolExecutor deadlocks
+        is_test_mode = os.environ.get("TEST_MODE") == "true"
+        if is_test_mode:
+            # Sequential processing in test mode - avoids ThreadPoolExecutor deadlocks
+            import contextlib
+
+            for feature, file_path in hash_tasks:
+                with contextlib.suppress(Exception):
+                    update_file_hash(feature, file_path)
+        else:
+            max_workers = max(1, min(multiprocessing.cpu_count() or 4, 16, len(hash_tasks)))
+            executor = ThreadPoolExecutor(max_workers=max_workers)
+            interrupted = False
             try:
-                for future in as_completed(future_to_task):
-                    try:
-                        future.result()
-                    except KeyboardInterrupt:
-                        interrupted = True
-                        for f in future_to_task:
-                            if not f.done():
-                                f.cancel()
-                        break
-                    except Exception:
-                        pass
+                future_to_task = {
+                    executor.submit(update_file_hash, feature, file_path): (feature, file_path)
+                    for feature, file_path in hash_tasks
+                }
+                try:
+                    for future in as_completed(future_to_task):
+                        try:
+                            future.result()
+                        except KeyboardInterrupt:
+                            interrupted = True
+                            for f in future_to_task:
+                                if not f.done():
+                                    f.cancel()
+                            break
+                        except Exception:
+                            pass
+                except KeyboardInterrupt:
+                    interrupted = True
+                    for f in future_to_task:
+                        if not f.done():
+                            f.cancel()
+                if interrupted:
+                    raise KeyboardInterrupt
             except KeyboardInterrupt:
                 interrupted = True
-                for f in future_to_task:
-                    if not f.done():
-                        f.cancel()
-            if interrupted:
-                raise KeyboardInterrupt
-        except KeyboardInterrupt:
-            interrupted = True
-            executor.shutdown(wait=False, cancel_futures=True)
-            raise
-        finally:
-            if not interrupted:
-                executor.shutdown(wait=True)
-            else:
-                executor.shutdown(wait=False)
+                executor.shutdown(wait=False, cancel_futures=True)
+                raise
+            finally:
+                if not interrupted:
+                    executor.shutdown(wait=True)
+                else:
+                    executor.shutdown(wait=False)
 
     for feature in plan_bundle.features:
         if feature.source_tracking:
@@ -441,6 +444,7 @@ def _extract_contracts(
     record_event: Any,
 ) -> dict[str, dict[str, Any]]:
     """Extract OpenAPI contracts from features."""
+    import os
     from concurrent.futures import ThreadPoolExecutor, as_completed
 
     from specfact_cli.generators.openapi_extractor import OpenAPIExtractor
@@ -474,45 +478,60 @@ def _extract_contracts(
 
         features_with_contracts = [f for f in plan_bundle.features if f.contract]
         if features_with_contracts:
-            max_workers = max(1, min(multiprocessing.cpu_count() or 4, 16, len(features_with_contracts)))
-            executor = ThreadPoolExecutor(max_workers=max_workers)
-            interrupted = False
+            import os
+
+            # In test mode, use sequential processing to avoid ThreadPoolExecutor deadlocks
+            is_test_mode = os.environ.get("TEST_MODE") == "true"
             existing_contracts_count = 0
-            try:
-                future_to_feature = {
-                    executor.submit(load_contract, feature): feature for feature in features_with_contracts
-                }
+            if is_test_mode:
+                # Sequential processing in test mode - avoids ThreadPoolExecutor deadlocks
+                for feature in features_with_contracts:
+                    try:
+                        feature_key, contract_data = load_contract(feature)
+                        if contract_data:
+                            contracts_data[feature_key] = contract_data
+                            existing_contracts_count += 1
+                    except Exception:
+                        pass
+            else:
+                max_workers = max(1, min(multiprocessing.cpu_count() or 4, 16, len(features_with_contracts)))
+                executor = ThreadPoolExecutor(max_workers=max_workers)
+                interrupted = False
                 try:
-                    for future in as_completed(future_to_feature):
-                        try:
-                            feature_key, contract_data = future.result()
-                            if contract_data:
-                                contracts_data[feature_key] = contract_data
-                                existing_contracts_count += 1
-                        except KeyboardInterrupt:
-                            interrupted = True
-                            for f in future_to_feature:
-                                if not f.done():
-                                    f.cancel()
-                            break
-                        except Exception:
-                            pass
+                    future_to_feature = {
+                        executor.submit(load_contract, feature): feature for feature in features_with_contracts
+                    }
+                    try:
+                        for future in as_completed(future_to_feature):
+                            try:
+                                feature_key, contract_data = future.result()
+                                if contract_data:
+                                    contracts_data[feature_key] = contract_data
+                                    existing_contracts_count += 1
+                            except KeyboardInterrupt:
+                                interrupted = True
+                                for f in future_to_feature:
+                                    if not f.done():
+                                        f.cancel()
+                                break
+                            except Exception:
+                                pass
+                    except KeyboardInterrupt:
+                        interrupted = True
+                        for f in future_to_feature:
+                            if not f.done():
+                                f.cancel()
+                    if interrupted:
+                        raise KeyboardInterrupt
                 except KeyboardInterrupt:
                     interrupted = True
-                    for f in future_to_feature:
-                        if not f.done():
-                            f.cancel()
-                if interrupted:
-                    raise KeyboardInterrupt
-            except KeyboardInterrupt:
-                interrupted = True
-                executor.shutdown(wait=False, cancel_futures=True)
-                raise
-            finally:
-                if not interrupted:
-                    executor.shutdown(wait=True)
-                else:
-                    executor.shutdown(wait=False)
+                    executor.shutdown(wait=False, cancel_futures=True)
+                    raise
+                finally:
+                    if not interrupted:
+                        executor.shutdown(wait=True)
+                    else:
+                        executor.shutdown(wait=False)
 
             if existing_contracts_count > 0:
                 console.print(f"[green]✓[/green] Loaded {existing_contracts_count} existing contract(s) from bundle")
@@ -527,10 +546,21 @@ def _extract_contracts(
         features_with_files = []
 
     if features_with_files and should_regenerate_contracts:
-        max_workers = max(1, min(multiprocessing.cpu_count() or 4, 16, len(features_with_files)))
-        console.print(
-            f"[cyan]📋 Extracting contracts from {len(features_with_files)} features (using {max_workers} workers)...[/cyan]"
-        )
+        import os
+
+        # In test mode, use sequential processing to avoid ThreadPoolExecutor deadlocks
+        is_test_mode = os.environ.get("TEST_MODE") == "true"
+        # Define max_workers for non-test mode (always defined to satisfy type checker)
+        max_workers = 1
+        if is_test_mode:
+            console.print(
+                f"[cyan]📋 Extracting contracts from {len(features_with_files)} features (sequential mode)...[/cyan]"
+            )
+        else:
+            max_workers = max(1, min(multiprocessing.cpu_count() or 4, 16, len(features_with_files)))
+            console.print(
+                f"[cyan]📋 Extracting contracts from {len(features_with_files)} features (using {max_workers} workers)...[/cyan]"
+            )
 
         from rich.progress import BarColumn, Progress, SpinnerColumn, TextColumn, TimeElapsedColumn
 
@@ -577,49 +607,67 @@ def _extract_contracts(
             console=console,
         ) as progress:
             task = progress.add_task("[cyan]Extracting contracts...", total=len(features_with_files))
-            executor = ThreadPoolExecutor(max_workers=max_workers)
-            interrupted = False
-            try:
-                future_to_feature = {executor.submit(process_feature, f): f for f in features_with_files}
+            if is_test_mode:
+                # Sequential processing in test mode - avoids ThreadPoolExecutor deadlocks
                 completed_count = 0
+                for feature in features_with_files:
+                    try:
+                        feature_key, openapi_spec = process_feature(feature)
+                        completed_count += 1
+                        progress.update(task, completed=completed_count)
+                        if openapi_spec:
+                            contract_ref = f"contracts/{feature_key}.openapi.yaml"
+                            feature.contract = contract_ref
+                            contracts_data[feature_key] = openapi_spec
+                            contracts_generated += 1
+                    except Exception as e:
+                        completed_count += 1
+                        progress.update(task, completed=completed_count)
+                        console.print(f"[dim]⚠ Warning: Failed to process feature: {e}[/dim]")
+            else:
+                executor = ThreadPoolExecutor(max_workers=max_workers)
+                interrupted = False
                 try:
-                    for future in as_completed(future_to_feature):
-                        try:
-                            feature_key, openapi_spec = future.result()
-                            completed_count += 1
-                            progress.update(task, completed=completed_count)
-                            if openapi_spec:
-                                feature = next(f for f in features_with_files if f.key == feature_key)
-                                contract_ref = f"contracts/{feature_key}.openapi.yaml"
-                                feature.contract = contract_ref
-                                contracts_data[feature_key] = openapi_spec
-                                contracts_generated += 1
-                        except KeyboardInterrupt:
-                            interrupted = True
-                            for f in future_to_feature:
-                                if not f.done():
-                                    f.cancel()
-                            break
-                        except Exception as e:
-                            completed_count += 1
-                            progress.update(task, completed=completed_count)
-                            console.print(f"[dim]⚠ Warning: Failed to process feature: {e}[/dim]")
+                    future_to_feature = {executor.submit(process_feature, f): f for f in features_with_files}
+                    completed_count = 0
+                    try:
+                        for future in as_completed(future_to_feature):
+                            try:
+                                feature_key, openapi_spec = future.result()
+                                completed_count += 1
+                                progress.update(task, completed=completed_count)
+                                if openapi_spec:
+                                    feature = next(f for f in features_with_files if f.key == feature_key)
+                                    contract_ref = f"contracts/{feature_key}.openapi.yaml"
+                                    feature.contract = contract_ref
+                                    contracts_data[feature_key] = openapi_spec
+                                    contracts_generated += 1
+                            except KeyboardInterrupt:
+                                interrupted = True
+                                for f in future_to_feature:
+                                    if not f.done():
+                                        f.cancel()
+                                break
+                            except Exception as e:
+                                completed_count += 1
+                                progress.update(task, completed=completed_count)
+                                console.print(f"[dim]⚠ Warning: Failed to process feature: {e}[/dim]")
+                    except KeyboardInterrupt:
+                        interrupted = True
+                        for f in future_to_feature:
+                            if not f.done():
+                                f.cancel()
+                    if interrupted:
+                        raise KeyboardInterrupt
                 except KeyboardInterrupt:
                     interrupted = True
-                    for f in future_to_feature:
-                        if not f.done():
-                            f.cancel()
-                if interrupted:
-                    raise KeyboardInterrupt
-            except KeyboardInterrupt:
-                interrupted = True
-                executor.shutdown(wait=False, cancel_futures=True)
-                raise
-            finally:
-                if not interrupted:
-                    executor.shutdown(wait=True)
-                else:
-                    executor.shutdown(wait=False)
+                    executor.shutdown(wait=False, cancel_futures=True)
+                    raise
+                finally:
+                    if not interrupted:
+                        executor.shutdown(wait=True)
+                    else:
+                        executor.shutdown(wait=False)
 
     elif should_regenerate_contracts:
         console.print("[dim]No features with implementation files found for contract extraction[/dim]")
@@ -740,14 +788,83 @@ def _save_bundle_if_needed(
     if should_regenerate_bundle:
         console.print("\n[cyan]💾 Compiling and saving project bundle...[/cyan]")
         project_bundle = _convert_plan_bundle_to_project_bundle(plan_bundle, bundle)
-        save_project_bundle(project_bundle, bundle_dir, atomic=True)
-        console.print("[green]✓[/green] Project bundle saved")
+        save_bundle_with_progress(project_bundle, bundle_dir, atomic=True, console_instance=console)
     else:
         console.print("\n[dim]⏭ Skipping bundle save (no changes detected)[/dim]")
 
 
-def _validate_api_specs(repo: Path) -> None:
-    """Validate OpenAPI/AsyncAPI specs with Specmatic if available."""
+def _validate_bundle_contracts(bundle_dir: Path, plan_bundle: PlanBundle) -> tuple[int, int]:
+    """
+    Validate OpenAPI/AsyncAPI contracts in bundle with Specmatic if available.
+
+    Args:
+        bundle_dir: Path to bundle directory
+        plan_bundle: Plan bundle containing features with contract references
+
+    Returns:
+        Tuple of (validated_count, failed_count)
+    """
+    import asyncio
+
+    from specfact_cli.integrations.specmatic import check_specmatic_available, validate_spec_with_specmatic
+
+    is_available, _error_msg = check_specmatic_available()
+    if not is_available:
+        return 0, 0
+
+    validated_count = 0
+    failed_count = 0
+    contract_files = []
+
+    # Collect contract files from features
+    # PlanBundle.features is a list, not a dict
+    features_iter = plan_bundle.features.values() if isinstance(plan_bundle.features, dict) else plan_bundle.features
+    for feature in features_iter:
+        if feature.contract:
+            contract_path = bundle_dir / feature.contract
+            if contract_path.exists():
+                contract_files.append((contract_path, feature.key))
+
+    if not contract_files:
+        return 0, 0
+
+    console.print(f"\n[cyan]🔍 Validating {len(contract_files)} contract(s) in bundle with Specmatic...[/cyan]")
+    for contract_path, feature_key in contract_files[:5]:  # Validate up to 5 contracts
+        console.print(f"[dim]Validating {contract_path.relative_to(bundle_dir)} (from {feature_key})...[/dim]")
+        try:
+            result = asyncio.run(validate_spec_with_specmatic(contract_path))
+            if result.is_valid:
+                console.print(f"  [green]✓[/green] {contract_path.name} is valid")
+                validated_count += 1
+            else:
+                console.print(f"  [yellow]⚠[/yellow] {contract_path.name} has validation issues")
+                if result.errors:
+                    for error in result.errors[:2]:
+                        console.print(f"    - {error}")
+                failed_count += 1
+        except Exception as e:
+            console.print(f"  [yellow]⚠[/yellow] Validation error: {e!s}")
+            failed_count += 1
+
+    if len(contract_files) > 5:
+        console.print(
+            f"[dim]... and {len(contract_files) - 5} more contract(s) (run 'specfact spec validate' to validate all)[/dim]"
+        )
+
+    return validated_count, failed_count
+
+
+def _validate_api_specs(repo: Path, bundle_dir: Path | None = None, plan_bundle: PlanBundle | None = None) -> None:
+    """
+    Validate OpenAPI/AsyncAPI specs with Specmatic if available.
+
+    Validates both repo-level spec files and bundle contracts if provided.
+
+    Args:
+        repo: Repository path
+        bundle_dir: Optional bundle directory path
+        plan_bundle: Optional plan bundle for contract validation
+    """
     import asyncio
 
     spec_files = []
@@ -761,8 +878,16 @@ def _validate_api_specs(repo: Path) -> None:
     ]:
         spec_files.extend(repo.glob(pattern))
 
+    validated_contracts = 0
+    failed_contracts = 0
+
+    # Validate bundle contracts if provided
+    if bundle_dir and plan_bundle:
+        validated_contracts, failed_contracts = _validate_bundle_contracts(bundle_dir, plan_bundle)
+
+    # Validate repo-level spec files
     if spec_files:
-        console.print(f"\n[cyan]🔍 Found {len(spec_files)} API specification file(s)[/cyan]")
+        console.print(f"\n[cyan]🔍 Found {len(spec_files)} API specification file(s) in repository[/cyan]")
         from specfact_cli.integrations.specmatic import check_specmatic_available, validate_spec_with_specmatic
 
         is_available, error_msg = check_specmatic_available()
@@ -787,6 +912,9 @@ def _validate_api_specs(repo: Path) -> None:
             console.print("[dim]💡 Tip: Run 'specfact spec mock' to start a mock server for development[/dim]")
         else:
             console.print(f"[dim]💡 Tip: Install Specmatic to validate API specs: {error_msg}[/dim]")
+    elif validated_contracts > 0 or failed_contracts > 0:
+        # Only show mock server tip if we validated contracts
+        console.print("[dim]💡 Tip: Run 'specfact spec mock' to start a mock server for development[/dim]")
 
 
 def _suggest_constitution_bootstrap(repo: Path) -> None:
@@ -1120,6 +1248,7 @@ def from_bridge(
         with Progress(
             SpinnerColumn(),
             TextColumn("[progress.description]{task.description}"),
+            TimeElapsedColumn(),
             console=console,
         ) as progress:
             # Step 1: Discover features from markdown artifacts
@@ -1199,7 +1328,7 @@ def from_bridge(
             project_bundle = _convert_plan_bundle_to_project_bundle(plan_bundle, bundle_name)
             bundle_dir = SpecFactStructure.project_dir(base_path=repo, bundle_name=bundle_name)
             SpecFactStructure.ensure_project_structure(base_path=repo, bundle_name=bundle_name)
-            save_project_bundle(project_bundle, bundle_dir, atomic=True)
+            save_bundle_with_progress(project_bundle, bundle_dir, atomic=True, console_instance=console)
             console.print(f"[dim]Project bundle: .specfact/projects/{bundle_name}/[/dim]")
 
         console.print("[bold green]✓[/bold green] Import complete!")
@@ -1470,8 +1599,8 @@ def from_code(
             console.print("\n[bold green]✓ Import complete![/bold green]")
             console.print(f"[dim]Project bundle written to: {bundle_dir}[/dim]")
 
-            # Validate API specs
-            _validate_api_specs(repo)
+            # Validate API specs (both repo-level and bundle contracts)
+            _validate_api_specs(repo, bundle_dir=bundle_dir, plan_bundle=plan_bundle)
 
             # Suggest constitution bootstrap
             _suggest_constitution_bootstrap(repo)
