@@ -40,6 +40,7 @@ class SpecFactStructure:
     GATES_RESULTS = f"{ROOT}/gates/results"
     CACHE = f"{ROOT}/cache"
     SDD = f"{ROOT}/sdd"  # SDD manifests (one per project bundle)
+    TASKS = f"{ROOT}/tasks"  # Task breakdowns (one per project bundle)
     CONFIG = f"{ROOT}/config"  # Global configuration (bridge.yaml, etc.)
 
     # Configuration files
@@ -248,6 +249,46 @@ class SpecFactStructure:
 
         return default_path
 
+    @classmethod
+    @beartype
+    @require(lambda base_path: base_path is None or isinstance(base_path, Path), "Base path must be None or Path")
+    @ensure(lambda result: result is None or isinstance(result, str), "Must return None or string")
+    def get_active_bundle_name(cls, base_path: Path | None = None) -> str | None:
+        """
+        Get active bundle name from config.
+
+        Args:
+            base_path: Base directory (default: current directory)
+
+        Returns:
+            Active bundle name (e.g., "main", "legacy-api") or None if not set
+        """
+        if base_path is None:
+            base_path = Path(".")
+        else:
+            base_path = Path(base_path).resolve()
+            parts = base_path.parts
+            if ".specfact" in parts:
+                specfact_idx = parts.index(".specfact")
+                base_path = Path(*parts[:specfact_idx])
+
+        config_path = base_path / cls.PLANS_CONFIG
+        if config_path.exists():
+            try:
+                import yaml
+
+                with config_path.open() as f:
+                    config = yaml.safe_load(f) or {}
+                active_plan = config.get("active_plan")
+                if active_plan:
+                    # Active plan is stored as bundle name (not plan filename)
+                    return active_plan
+            except Exception:
+                # Fallback to None if config read fails
+                pass
+
+        return None
+
     @classmethod
     @beartype
     @require(lambda base_path: base_path is None or isinstance(base_path, Path), "Base path must be None or Path")
